"""Module that contains tools class.

Attributes:
    log (logging.Log): current logger
"""
import subprocess
import logging

log = logging.getLogger('fetch')


class Tools(object):
    """Some useful tools.

    Attributes:
        default_ros_packages (str[]): default packages for ROS
    """

    @staticmethod
    def prepare_default_url(default_url):
        """Insert {package} statement into the URL.

        Args:
            default_url (str): Defatul url. Can be one of the following styles:
                - git@<path>
                - https://<path>

        Returns:
            str: Url with {package} tag placed inside of it.
        """
        package_tag = '{package}'
        if len(default_url) == 0:
            return default_url
        if not default_url.endswith('/'):
            default_url += '/'
        if default_url.startswith('git'):
            return default_url + package_tag + '.git'
        if default_url.startswith('http'):
            return default_url + package_tag
        return default_url

    @staticmethod
    def list_all_ros_pkgs():
        """List all available ROS packages.

        Returns:
            str[]: list of ROS package names
        """
        log.info(" Avoid fetching ROS packages.")
        log.info(" [ROS]: Searching all packages.")
        get_ros_packages_command = 'rospack list'
        pkg_list = []
        try:
            output = subprocess.check_output(
                get_ros_packages_command,
                stderr=subprocess.STDOUT,
                shell=True)
            output = output.splitlines()
            for pkg_line in output:
                pkg_list.append(pkg_line.split(' ')[0])
            log.info(" [ROS]: Ignoring %s packages.", len(pkg_list))
            return set(pkg_list)
        except subprocess.CalledProcessError:
            log.info(" [ROS]: Not found. Ignoring pre-defined ROS packages.")
            return set(Tools.default_ros_packages)

    @staticmethod
    def decorate(pkg_name):
        """Decorate a package name."""
        return "[" + pkg_name + "]"

    default_ros_packages = ['actionlib',
                            'actionlib_msgs',
                            'actionlib_tutorials',
                            'angles',
                            'bond',
                            'bondcpp',
                            'bondpy',
                            'camera_calibration',
                            'camera_calibration_parsers',
                            'camera_info_manager',
                            'catkin',
                            'class_loader',
                            'cmake_modules',
                            'collada_parser',
                            'collada_urdf',
                            'compressed_depth_image_transport',
                            'compressed_image_transport',
                            'control_msgs',
                            'control_toolbox',
                            'controller_interface',
                            'controller_manager',
                            'controller_manager_msgs',
                            'cpp_common',
                            'cv_bridge',
                            'depth_image_proc',
                            'diagnostic_aggregator',
                            'diagnostic_analysis',
                            'diagnostic_common_diagnostics',
                            'diagnostic_msgs',
                            'diagnostic_updater',
                            'driver_base',
                            'dynamic_reconfigure',
                            'eigen_conversions',
                            'eigen_stl_containers',
                            'filters',
                            'gazebo_msgs',
                            'gazebo_plugins',
                            'gazebo_ros',
                            'gazebo_ros_control',
                            'gencpp',
                            'genlisp',
                            'genmsg',
                            'genpy',
                            'geometric_shapes',
                            'geometry_msgs',
                            'hardware_interface',
                            'image_geometry',
                            'image_proc',
                            'image_rotate',
                            'image_transport',
                            'image_view',
                            'interactive_marker_tutorials',
                            'interactive_markers',
                            'joint_limits_interface',
                            'joint_state_publisher',
                            'kdl_conversions',
                            'kdl_parser',
                            'laser_assembler',
                            'laser_filters',
                            'laser_geometry',
                            'librviz_tutorial',
                            'map_msgs',
                            'media_export',
                            'message_filters',
                            'message_generation',
                            'message_runtime',
                            'mk',
                            'nav_msgs',
                            'nodelet',
                            'nodelet_topic_tools',
                            'nodelet_tutorial_math',
                            'octomap',
                            'opencv_apps',
                            'openni2_camera',
                            'openni2_launch',
                            'orocos_kdl',
                            'pcl_conversions',
                            'pcl_msgs',
                            'pcl_ros',
                            'pluginlib',
                            'pluginlib_tutorials',
                            'pointcloud_to_laserscan',
                            'polled_camera',
                            'pr2_description',
                            'python_orocos_kdl',
                            'python_qt_binding',
                            'qt_dotgraph',
                            'qt_gui',
                            'qt_gui_cpp',
                            'qt_gui_py_common',
                            'random_numbers',
                            'realtime_tools',
                            'resource_retriever',
                            'rgbd_launch',
                            'robot_state_publisher',
                            'rosbag',
                            'rosbag_migration_rule',
                            'rosbag_storage',
                            'rosbash',
                            'rosboost_cfg',
                            'rosbuild',
                            'rosclean',
                            'rosconsole',
                            'rosconsole_bridge',
                            'roscpp',
                            'roscpp_serialization',
                            'roscpp_traits',
                            'roscpp_tutorials',
                            'roscreate',
                            'rosgraph',
                            'rosgraph_msgs',
                            'roslang',
                            'roslaunch',
                            'roslib',
                            'roslint',
                            'roslisp',
                            'roslz4',
                            'rosmake',
                            'rosmaster',
                            'rosmsg',
                            'rosnode',
                            'rosout',
                            'rospack',
                            'rosparam',
                            'rospy',
                            'rospy_tutorials',
                            'rosserial_arduino',
                            'rosserial_client',
                            'rosserial_msgs',
                            'rosserial_python',
                            'rosservice',
                            'rostest',
                            'rostime',
                            'rostopic',
                            'rosunit',
                            'roswtf',
                            'rqt_action',
                            'rqt_bag',
                            'rqt_bag_plugins',
                            'rqt_console',
                            'rqt_dep',
                            'rqt_graph',
                            'rqt_gui',
                            'rqt_gui_cpp',
                            'rqt_gui_py',
                            'rqt_image_view',
                            'rqt_launch',
                            'rqt_logger_level',
                            'rqt_moveit',
                            'rqt_msg',
                            'rqt_nav_view',
                            'rqt_plot',
                            'rqt_pose_view',
                            'rqt_publisher',
                            'rqt_py_common',
                            'rqt_py_console',
                            'rqt_reconfigure',
                            'rqt_robot_dashboard',
                            'rqt_robot_monitor',
                            'rqt_robot_steering',
                            'rqt_runtime_monitor',
                            'rqt_rviz',
                            'rqt_service_caller',
                            'rqt_shell',
                            'rqt_srv',
                            'rqt_tf_tree',
                            'rqt_top',
                            'rqt_topic',
                            'rqt_web',
                            'rviz',
                            'rviz_plugin_tutorials',
                            'rviz_python_tutorial',
                            'self_test',
                            'sensor_msgs',
                            'shape_msgs',
                            'smach',
                            'smach_msgs',
                            'smach_ros',
                            'smclib',
                            'stage',
                            'stage_ros',
                            'std_msgs',
                            'std_srvs',
                            'stereo_image_proc',
                            'stereo_msgs',
                            'tf',
                            'tf2',
                            'tf2_geometry_msgs',
                            'tf2_kdl',
                            'tf2_msgs',
                            'tf2_py',
                            'tf2_ros',
                            'tf2_sensor_msgs',
                            'tf_conversions',
                            'theora_image_transport',
                            'topic_tools',
                            'trajectory_msgs',
                            'transmission_interface',
                            'turtle_actionlib',
                            'turtle_tf',
                            'turtle_tf2',
                            'turtlesim',
                            'urdf',
                            'urdf_parser_plugin',
                            'urdf_tutorial',
                            'urdfdom_py',
                            'velodyne_driver',
                            'velodyne_msgs',
                            'velodyne_pointcloud',
                            'visualization_marker_tutorials',
                            'visualization_msgs',
                            'xacro',
                            'xmlrpcpp']



"""Contains main for deps verb.

Attributes:
    log (logging.Log): logger
"""

import sys
import logging
from os import path

try:
    from catkin_pkg.packages import find_packages
except ImportError as e:
    sys.exit(
        'ImportError: "from catkin_pkg.topological_order import '
        'topological_order" failed: %s\nMake sure that you have installed '
        '"catkin_pkg", and that it is up to date and on the PYTHONPATH.' % e
    )

from catkin_tools.argument_parsing import add_context_args
from catkin_tools.context import Context

from catkin_tools_fetch.fetcher.dependency_parser import Parser
from catkin_tools_fetch.fetcher.downloader import Downloader
from catkin_tools_fetch.fetcher.tools import Tools

logging.basicConfig()
log = logging.getLogger('fetch')


def prepare_arguments(parser):
    """Parse arguments that belong to this verb.

    Args:
        parser (argparser): Argument parser

    Returns:
        argparser: Parser that knows about our flags.
    """
    parser.description = """ Download dependencies for one or more packages in
        a catkin workspace. This reads dependencies from package.xml file of
        each of the packages in the workspace and tries to download their
        sources from version control system of choice."""
    add_context_args(parser)

    packages_help_msg = """
        Packages for which the dependencies are analyzed.
        If no packages are given, all packages are processed."""
    fetch_group = parser.add_argument_group(
        'Packages',
        'Control for which packages we fetch dependencies.')

    fetch_group.add_argument('packages',
                             metavar='PKGNAME',
                             nargs='*',
                             help=packages_help_msg)

    # add config flags to all groups that need it
    config_group = parser.add_argument_group('Config')
    config_group.add_argument(
        '--default_url', default="{package}",
        help='Where to look for packages by default.')
    config_group.add_argument(
        '--update', '-u', action='store_true', default=False,
        help='Update the dependencies to latest version.')

    # Behavior
    behavior_group = parser.add_argument_group(
        'Interface', 'The behavior of the command-line interface.')
    add = behavior_group.add_argument
    add('--verbose', '-v', action='store_true', default=False,
        help='Print output from commands.')

    return parser


def main(opts):
    """Main function for fetch verb.

    Args:
        opts (dict): Options populated by an arg parser.

    Returns:
        int: Return code
    """
    # Load the context
    if opts.verbose:
        log.setLevel(logging.getLevelName("DEBUG"))
        log.debug(" Enabling DEBUG output.")
    else:
        log.setLevel(logging.getLevelName("INFO"))

    context = Context.load(opts.workspace, opts.profile, opts, append=True)
    default_url = Tools.prepare_default_url(opts.default_url)

    if not opts.workspace:
        log.critical(" Workspace undefined! Abort!")
        return 1
    if opts.update:
        log.error(" Sorry, 'update' not implemented yet, but is planned.")
        return 1
    if opts.verb == 'fetch':
        return fetch(packages=opts.packages,
                     workspace=opts.workspace,
                     context=context,
                     default_url=default_url)


def fetch(packages, workspace, context, default_url):
    """Detch dependencies of a package.

    Args:
        packages (list): A list of packages provided by the user.
        workspace (str): Path to a workspace (without src/ in the end).
        context (Context): Current context. Needed to find current packages.
        default_url (str): A default url with a {package} placeholder in it.

    Returns:
        int: Return code. 0 if success. Git error code otherwise.
    """
    fetch_all = False
    if len(packages) == 0:
        fetch_all = True

    ws_path = path.join(workspace, 'src')
    ignore_pkgs = Tools.list_all_ros_pkgs()

    already_fetched = set()

    global_error_code = Downloader.NO_ERROR

    # loop until there are still any new dependencies left to download
    while(True):
        log.info(" Searching for dependencies.")
        deps_to_fetch = {}
        workspace_packages = find_packages(
            context.source_space_abs,
            exclude_subspaces=True, warnings=[])
        available_pkgs = [pkg.name for _, pkg in workspace_packages.items()]
        initial_cloned_pkgs = len(already_fetched)
        for package_path, package in workspace_packages.items():
            if package.name in already_fetched:
                continue
            if fetch_all or (package.name in packages):
                parser = Parser(default_url, package.name)
                package_folder = path.join(ws_path, package_path)
                deps_to_fetch.update(parser.get_dependencies(package_folder))
                already_fetched.add(package.name)
        try:
            downloader = Downloader(ws_path, available_pkgs, ignore_pkgs)
        except ValueError as e:
            log.critical(" Encountered error. Abort.")
            log.critical(" Error message: %s", e.message)
            return 1
        error_code = downloader.download_dependencies(deps_to_fetch)
        if len(already_fetched) == initial_cloned_pkgs:
            log.info(" No new dependencies. Done.")
            break
        if error_code != 0:
            global_error_code = error_code
        log.info(" New packages available. Process their dependencies now.")
    return global_error_code



"""Encapsulates a dependency parser.

Attributes:
    log (logging.Log): logger
"""
import os
import logging
from os import path
from xml.dom import minidom

from catkin_tools_fetch.fetcher.tools import Tools

log = logging.getLogger('fetch')


class Parser(object):
    """Parses dependencies of a package.

    Attributes:
        pkg_name (str): Name of currently parsed package.
        TAGS (list): A list of tags that we want to parse.
        URL_TAGS (list): A list of tags to consider when parsing explicit urls.
        XML_FILE_NAME (str): A generic xml name of package to be parsed.
    """

    XML_FILE_NAME = "package.xml"
    TAGS = ["build_depend"]
    URL_TAGS = ["git_url"]

    def __init__(self, download_mask, pkg_name):
        """Initialize a dependency parser.

        Args:
            download_mask (str): a mask containing {package} tag to be replaced
                later, e.g. git@<path>/{package}.git
            pkg_name (str): Name of current package
        """
        super(Parser, self).__init__()
        if '{package}' not in download_mask:
            raise ValueError(
                '`download_mask` must contain a "{package}" placeholder.')
        self.__download_mask = download_mask
        self.pkg_name = pkg_name

    def get_dependencies(self, package_folder):
        """Find and parse package.xml file and return a dict of dependencies.

        Args:
            package_folder (str): A folder to search package.xml in.

        Returns:
            dict: A dictionary with an url for each package name.
        """
        path_to_xml = Parser.__get_package_xml_path(package_folder)
        if not path_to_xml:
            log.critical(" 'package.xml' not found for package [%s].",
                         self.pkg_name)
            return None
        return self.__get_all_dependencies(path_to_xml)

    def __get_all_dependencies(self, path_to_xml):
        """Get a dictionary of all dependencies.

        Args:
            path_to_xml (str): Path to `package.xml` file.

        Returns:
            dict: Dictionary with an url for each dependency name.
        """
        xmldoc = minidom.parse(path_to_xml)
        all_deps = []
        for tag in Parser.TAGS:
            deps = Parser.__node_to_list(xmldoc, tag)
            deps = Parser.__fix_dependencies(deps, self.pkg_name)
            all_deps += deps
        log.info("  %-21s: Found %s dependencies.",
                 Tools.decorate(self.pkg_name),
                 len(all_deps))
        log.debug(" Dependencies: %s", all_deps)
        deps_with_urls = self.__init_dep_dict(all_deps)
        return Parser.__specify_explicit_urls(xmldoc, deps_with_urls)

    @staticmethod
    def __fix_dependencies(deps, pkg_name):
        """Fix dependencies if they are malformed.

        Args:
            deps (str[]): List of dependencies.
            pkg_name (str): Current package name.

        Returns:
            str[]: Fixed dependencies.
        """
        fixed_deps = list(deps)
        for i, dep in enumerate(deps):
            fixed_deps[i] = dep.strip()
            if fixed_deps[i] != dep:
                log.warning(
                    " [%s]: Fix dependency in `package.xml`: [%s] to [%s]",
                    pkg_name, dep, fixed_deps[i])
        return fixed_deps

    @staticmethod
    def __specify_explicit_urls(xmldoc, intial_dep_dict):
        """Specify explicit urls instead of default ones.

        Args:
            xmldoc (minidom): Current xml object
            intial_dep_dict (dict): A dict {name: url} with default urls.

        Returns:
            dict: A dict with specific ursl parsed from <export> tags
        """
        dep_dict = dict(intial_dep_dict)
        for url_tag in Parser.URL_TAGS:
            urls_node = xmldoc.getElementsByTagName(url_tag)
            for item in urls_node:
                target = item.attributes['target'].value
                url = item.attributes['url'].value
                dep_dict[target] = url
        return dep_dict

    def __init_dep_dict(self, all_deps_list):
        """Initialize dependency dict with default values.

        Args:
            all_deps_list (str[]): List of all dependencies.

        Returns:
            dict: A dictionary {name: url} with default url for each dependency
        """
        dep_dict = {}
        for dependency in all_deps_list:
            dep_dict[dependency] = self.__download_mask.format(
                package=dependency)
        return dep_dict

    @staticmethod
    def __node_to_list(xmldoc, xml_node_name):
        """Get list of elements for a name of node.

        Args:
            xmldoc (minidom): Object for xml file.
            xml_node_name (str): Node name.

        Returns:
            str[]: List of all names in node
        """
        node = xmldoc.getElementsByTagName(xml_node_name)
        return [str(s.childNodes[0].nodeValue) for s in node]

    @staticmethod
    def __get_package_xml_path(folder):
        """Get path of `package.xml` file in folder.

        Args:
            folder (str): Folder to search file in.

        Returns:
            str: Full path to file or None if not found.
        """
        for file in os.listdir(folder):
            if file == Parser.XML_FILE_NAME:
                full_path = path.join(folder, file)
                return full_path
        return None



"""Hosts a class used to download dependencies.

Attributes:
    log (logging.Log): logger
"""
import logging
import subprocess

from os import path
from subprocess import PIPE

from catkin_tools_fetch.fetcher.tools import Tools

log = logging.getLogger('fetch')


class Downloader(object):
    """Downloader for dependency dict.

    Attributes:
        available_pkgs (str[]): dict of available packages in workspace
        ignore_pkgs (set): a set of packages to ignore (mostly ROS ones).
        ws_path (str): Workspace path. This is where packages live.
    """

    GIT_CHECK_CMD_MASK = "git ls-remote {url}"
    GIT_CLONE_CMD_MASK = "git clone --recursive {url} {path}"

    IGNORE_TAG = "[IGNORED]"
    NOT_FOUND_TAG = "[NOT FOUND]"

    EXISTS_TAG = "[ALREADY EXISTS]"
    CLONED_TAG = "[CLONED]"
    ERROR_TAG = "[ERROR]"

    NO_ERROR = 0

    def __init__(self, ws_path, available_pkgs, ignore_pkgs):
        """Init a downloader.

        Args:
            ws_path (str): Workspace path. This is where packages live.
            available_pkgs (iterable): dict of available packages in workspace.
            ignore_pkgs (iterable): set of packages to ignore (e.g. ROS ones).
        """
        super(Downloader, self).__init__()
        if not path.exists(ws_path):
            raise ValueError("""
            Folder '{}' is missing.
            Are you running 'fetch' from a catkin workspace?
            """.format(ws_path))
        self.ws_path = ws_path
        self.available_pkgs = available_pkgs
        self.ignore_pkgs = ignore_pkgs

    def download_dependencies(self, dep_dict):
        """Check and download dependencies from a dependency dictionary.

        Args:
            dep_dict (dict): dictionary {name: url} with links to repos.

        Returns:
            int: Return code. 0 if all fine. Git error code otherwise.
        """
        if not isinstance(dep_dict, dict):
            raise ValueError("expected a dictionary with dependencies.")
        checked_deps = self.__check_dependencies(dep_dict)
        return self.__clone_dependencies(checked_deps)

    def __clone_dependencies(self, checked_deps):
        """Clone dependencies.

        Args:
            checked_deps (dict): Dict {name: url} with valid urls to repos.

        Returns:
            int: Error code. 0 if all fine, Result from git error otherwise.
        """
        if not checked_deps:
            # exit early if there is nothing to download
            return Downloader.NO_ERROR
        log.info(" Cloning valid dependencies:")
        error_code = Downloader.NO_ERROR
        for name, url in checked_deps.items():
            if name in self.available_pkgs:
                log.info("  %-21s: %s",
                         Tools.decorate(name),
                         Downloader.EXISTS_TAG)
                continue
            dep_path = path.join(self.ws_path, name)
            cmd_clone = Downloader.GIT_CLONE_CMD_MASK.format(
                url=url, path=dep_path)
            try:
                subprocess.check_output(cmd_clone,
                                        stderr=subprocess.STDOUT,
                                        shell=True)
                log.info("  %-21s: %s",
                         Tools.decorate(name),
                         Downloader.CLONED_TAG)
            except subprocess.CalledProcessError as e:
                out_str = e.output.decode("utf8")
                log.debug(" Clone output: %s", out_str)
                if "already exists" in out_str:
                    log.info("  %-21s: %s",
                             Tools.decorate(name),
                             Downloader.EXISTS_TAG)
                else:
                    log.error("  %-21s: %s Git error code: %s",
                              Tools.decorate(name),
                              Downloader.ERROR_TAG,
                              e.returncode)
                    error_code = e.returncode
        return error_code

    def __check_dependencies(self, dep_dict):
        """Check dependencies for validity.

        Args:
            dep_dict (dict): A dictionary {name: url} with dependencies.

        Returns:
            dict: Only valid dependencies from the input dict.
        """
        checked_deps = {}
        if not dep_dict:
            # exit early if there are no new dependencies
            return checked_deps
        log.info(" Checking merged dependencies:")
        for name, url in dep_dict.items():
            if name in self.ignore_pkgs:
                log.info("  %-21s: %s",
                         Tools.decorate(name),
                         Downloader.IGNORE_TAG)
            elif Downloader.repository_exists(url):
                log.info("  %-21s: %s", Tools.decorate(name), url)
                checked_deps[name] = url
            else:
                log.info("  %-21s: %s",
                         Tools.decorate(name),
                         Downloader.NOT_FOUND_TAG)
        return checked_deps

    @staticmethod
    def repository_exists(url):
        """Check if repository exists.

        Uses `git ls-remote` to check if the repository exists.

        Args:
            url (str): Url to check.

        Returns:
            bool: True if exists, False otherwise
        """
        if url == "":
            return False
        git_cmd = Downloader.GIT_CHECK_CMD_MASK.format(url=url)
        try:
            subprocess.check_call(git_cmd, stdout=PIPE,
                                  stderr=PIPE, shell=True)
            return True
        except subprocess.CalledProcessError:
            return False



import os
import sys
from stat import ST_MODE
from distutils import log
from setuptools import setup
from setuptools import find_packages
from setuptools.command.install import install

# Setup installation dependencies
install_requires = [
    'catkin-pkg > 0.2.9',
    'catkin_tools >= 0.4.2',
    'setuptools',
]
if sys.version_info[0] == 2 and sys.version_info[1] <= 6:
    install_requires.append('argparse')

# Figure out the resources that need to be installed
this_dir = os.path.abspath(os.path.dirname(__file__))
osx_resources_path = os.path.join(
    this_dir,
    'catkin_tools',
    'notifications',
    'resources',
    'osx',
    'catkin build.app')


class PermissiveInstall(install):
    def run(self):
        install.run(self)
        if os.name == 'posix':
            for file in self.get_outputs():
                # all installed files should be readable for anybody
                mode = ((os.stat(file)[ST_MODE]) | 0o444) & 0o7777
                log.info("changing permissions of %s to %o" % (file, mode))
                os.chmod(file, mode)


version_str = '0.0.6'
github_url = 'https://github.com/niosus/catkin_tools_fetch'

setup(
    name='catkin_tools_fetch',
    packages=find_packages(exclude=['tests', 'docs']),
    version=version_str,
    install_requires=install_requires,
    author='Igor Bogoslavskyi',
    author_email='igor.bogoslavskyi@uni-bonn.de',
    maintainer='Igor Bogoslavskyi',
    maintainer_email='igor.bogoslavskyi@uni-bonn.de',
    keywords=['catkin', 'catkin_tools'],
    license="Apache 2.0",
    url=github_url,
    download_url=github_url + '/tarball/' + version_str,
    classifiers=[
        'Environment :: Console',
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
    ],
    description="A new verb 'fetch' for catkin_tools",
    long_description="""
Provides a new verb 'fetch' for catkin_tools. Allows fetching dependencies of
the packages found inside the catkin workspace.
""",
    test_suite='tests',
    entry_points={
        'catkin_tools.commands.catkin.verbs': [
            'fetch = catkin_tools_fetch:description',
        ],
    },
    cmdclass={'install': PermissiveInstall},
)
